"""
AST graphic representation Module.

This module provides utilities for converting an Abstract Syntax Tree (AST),
represented as a nested Python dictionary, to a Graphviz dot graph which
can be displayed inline in a Jupyter notebook, or as an ascii representation
directly in the console.
"""

from __future__ import annotations

import re
import types

from typing import Optional, cast

import requests

from asciinet import (
    Timeout,
    _asciigraph,
    _AsciiGraphProxy,
)
from graphviz import Digraph
from IPython.display import Image, display
from msgpack import dumps, loads

from astx.base import DictDataTypesStruct, ReprStruct


def traverse_ast_ascii(
    ast: ReprStruct,
    graph: Optional[Digraph] = None,
    parent: Optional[str] = None,
    shape: str = "box",
) -> Digraph:
    """
    Traverse the AST and build a Graphviz graph for ascii representation.

    Parameters
    ----------
    ast : dict
        The AST as a nested dictionary (full structure version).
    graph : Digraph
        The Graphviz graph object.
    parent : str, optional
        The identifier of the parent node in the graph, by default
        it is an empty string
    shape: str, options: ellipse, box, circle, diamond
        The shape used for the nodes in the graph. Default "box".

    Returns
    -------
    Digraph
        Graphviz (dot) graph representation.
    """
    if not graph:
        graph = Digraph()
        graph.attr(rankdir="TB")

    if isinstance(ast, list):
        for item in ast:
            traverse_ast_ascii(cast(ReprStruct, item), graph, parent, shape)
    elif isinstance(ast, dict):
        for key, value in ast.items():
            if not parent:
                node_name = f"{hash(key)}"
            else:
                if parent.find("_"):
                    node_name = f"{parent[parent.find('_')+1:]}_{hash(key)}"
                else:
                    node_name = f"{parent}_{hash(key)}"
                graph.edge(parent, node_name)

            graph.node(node_name, label=key, shape=shape)
            traverse_ast_ascii(
                cast(ReprStruct, value), graph, node_name, shape
            )
    return graph


def traverse_ast_to_graphviz(
    ast: ReprStruct,
    graph: Optional[Digraph] = None,
    parent: Optional[str] = None,
    shape: str = "box",
    edge_label: str = "",
) -> Digraph:
    """
    Traverse the AST and build a Graphviz graph for png representation.

    Parameters
    ----------
    ast : dict
        The AST as a nested dictionary (full structure version).
    graph : Digraph
        The Graphviz graph object.
    parent : str, optional
        The identifier of the parent node in the graph, by default
        it is an empty string
    shape: str, options: ellipse, box, circle, diamond
        The shape used for the nodes in the graph. Default "box".

    Returns
    -------
    Digraph
        Graphviz (dot) graph representation.
    """
    if not graph:
        graph = Digraph()
        graph.attr(rankdir="TB")

    if not isinstance(ast, dict):
        return graph.unflatten(stagger=3)

    for key, full_value in ast.items():
        if not isinstance(full_value, dict):
            continue

        content = full_value.get("content", "")
        metadata = cast(DictDataTypesStruct, full_value.get("metadata", {}))
        ref = ""

        if not metadata:
            # if the node doesn't have a metadata, it is a edge
            traverse_ast_to_graphviz(
                full_value,
                graph,
                parent,
                shape=shape,
                edge_label=key,
            )
            continue

        ref = cast(str, metadata.get("ref", ""))

        node_name = f"{hash(key)}_{hash(str(ref))}_{hash(str(content))}"
        graph.node(node_name, key, shape=shape)

        if parent:
            graph_params = {"label": edge_label} if edge_label else {}
            graph.edge(parent, node_name, **graph_params)

        if isinstance(content, dict):
            traverse_ast_to_graphviz(content, graph, node_name, shape=shape)
            continue
        elif not isinstance(content, list):
            continue

        for item in content:
            if isinstance(item, dict):
                traverse_ast_to_graphviz(item, graph, node_name, shape=shape)
    return graph


def visualize(ast: ReprStruct, shape: str = "box") -> None:
    """
    Visualize the abstract syntax tree (AST) using graphviz.

    Parameters
    ----------
    ast: dict
            The AST as a nested dictionary
    shape: str, options: ellipse, box, circle, diamond.
        The shape used for the nodes in the graph. Default "box".
    """
    graph = traverse_ast_to_graphviz(ast, shape=shape)
    image = Image(  # type: ignore[no-untyped-call]
        graph.unflatten(stagger=3).pipe(format="png")
    )
    display(image)  # type: ignore[no-untyped-call]


def make_node_box(modhash_label_mapping: list[tuple[str, str]]) -> str:
    """
    Make ascii representation for one-node ASTs.

    Parameters
    ----------
    modhash_label_mapping : list
        Mapping between node hash and label.

    Returns
    -------
    str
        The ascii graph representation as a string.

    """
    label = modhash_label_mapping[0][1]
    box_width = len(label) + 2
    space_before_box = " " * 4
    box_upper = space_before_box + "┌" + "─" * box_width + "┐"
    box_middle = space_before_box + "│ " + label + " │"
    box_lower = space_before_box + "└" + "─" * box_width + "┘"
    box = [box_upper, box_middle, box_lower]
    node = "\n".join(box)
    return node


def graph_to_ascii_overload(
    self: _AsciiGraphProxy, graph: Digraph, timeout: int = 10
) -> str:
    """
    Overload asciinet.graph_to_ascii function.

    Create an ascii representation of the abstract syntax tree (AST).
    This function is suitable for usage with ASTs with multiple nodes
    with the same label.

    Parameters
    ----------
    graph : Digraph
        The Graphviz graph object.
    timeout : int
        Time limit in seconds for requests.post. Default is 10 seconds.

    Returns
    -------
    str
        The ascii graph representation as a string.

    """
    try:
        nodes_modhash, edges_modhash, modhash_label_mapping = get_hash_labels(
            graph
        )

        # assuming there won't be more than one node with no edges
        if not edges_modhash:
            node = make_node_box(modhash_label_mapping)
            return node

        # Prepare the graph ascii repr
        graph_repr = dumps({"vertices": nodes_modhash, "edges": edges_modhash})
        response = requests.post(self._url, data=graph_repr, timeout=timeout)
        success = 200
        if response.status_code == success:
            graph_str = loads(response.content)
        else:
            raise ValueError(
                "Internal error: \n{0}".format(response.content.decode())
            )

        # substitute modhash by labels in the ascii representation
        graph_list = list(graph_str)
        for modhash, label in modhash_label_mapping:
            start = graph_str.index(modhash)
            end = graph_str.index(modhash) + (len(modhash))
            graph_list[start:end] = label

        graph = "".join(graph_list)
        return graph  # type: ignore[no-any-return]

    except (ConnectionError, Timeout):
        self._restart()
        raise ValueError("Could not convert graph to ASCII")


def get_hash_labels(
    graph: Digraph,
) -> tuple[list[str], list[tuple[str, str]], list[tuple[str, str]]]:
    """
    Get hash and labels from Digraph for ascii AST representation.

    Parameters
    ----------
    graph : Digraph
        The Graphviz graph object.

    Returns
    -------
    list
        Hash for ascii representation nodes.
    list
        Hash for ascii representation edges.
    list
        Mapping between nodes hash and labels.
    """
    dot_lines = graph.source.splitlines()
    nodes_modhash = []
    edges_modhash = []
    sources_hash = []
    targets_hash = []
    modhash_label_mapping = []
    hash_modhash_mapping = []

    for dot_line in dot_lines:
        line = dot_line.strip().strip(";")
        if "label" in line:
            node_label = re.findall(r"(?<=label=).*(?= )", line)[0].replace(
                '"', ""
            )

            # all labels must be at least 7 characters long
            node_label = node_label.center(7, " ")
            len_label = len(node_label)

            # each node modhash will have the same length as the node label
            # and will consist of parts of the hash from both the parent
            # (if it exists) and child nodes, separated by underscore.
            node_hash = line.split("[")[0].strip().replace('"', "")
            if "_" not in node_hash:  # if it's the first node
                if len_label <= len(node_hash):
                    x = len_label
                    node_modhash = node_hash[:x]
                else:
                    node_modhash = node_hash + " " * (
                        len_label - len(node_hash)
                    )
            else:  # if it's connected before and after
                hash1, hash2 = node_hash.split("_")
                len_hash2 = len(hash2)
                len_hash1 = len(hash1)
                min_chars_hash1 = 3

                standard_label_len = (min_chars_hash1 + len_hash2) + 1
                long_label_len = len_hash1 + len_hash2 + 1
                # short label:
                # modhash will have 3 chars of hash1 and some part of hash2
                if len_label <= standard_label_len:
                    nchars_hash2 = len_label - (min_chars_hash1 + 1)
                    node_modhash = (
                        f"{hash1[:min_chars_hash1]}_{hash2[:nchars_hash2]}"
                    )
                # medium label:
                # modhash will have more than 3 chars of hash1 and all of hash2
                elif (len_label > standard_label_len) & (
                    len_label <= long_label_len
                ):
                    nchars_hash1 = (
                        len_label - standard_label_len + min_chars_hash1
                    )
                    node_modhash = f"{hash1[:nchars_hash1]}_{hash2}"
                # long label:
                # modhash will have all of hash1, all of hash2,
                # plus some additional chars
                else:
                    nchars = len_label - long_label_len
                    add_chars = "x" * nchars
                    node_modhash = f"{hash1}_{hash2}{add_chars}"

            nodes_modhash.append(node_modhash)
            hash_modhash_mapping.append((node_hash, node_modhash))
            modhash_label_mapping.append((node_modhash, node_label))

        elif "->" in line:
            source_hash, target_hash = line.split("->")
            sources_hash.append(source_hash.strip().replace('"', ""))
            targets_hash.append(target_hash.strip().replace('"', ""))

    for source_hash, target_hash in zip(sources_hash, targets_hash):
        source_modhash = next(
            modhash
            for hash_, modhash in hash_modhash_mapping
            if source_hash == hash_
        )
        target_modhash = next(
            modhash
            for hash_, modhash in hash_modhash_mapping
            if target_hash == hash_
        )

        edges_modhash.append((source_modhash, target_modhash))

    return nodes_modhash, edges_modhash, modhash_label_mapping


def graph_to_ascii(graph: Digraph, timeout: int = 10) -> str:
    """
    Wrap function for graph_to_ascii.

    Create an ascii representation of the abstract syntax tree (AST).

    Parameters
    ----------
    graph : Digraph
        The Graphviz graph object.
    timeout : int
        Time limit in seconds for requests.post. Default is 10 seconds.
    """
    if not isinstance(graph, Digraph):
        raise ValueError(
            f"Graph must be a graphviz.Digraph (`{type(graph)}` was given.)"
        )

    result = _asciigraph.graph_to_ascii(graph, timeout=timeout)
    return f"\n{result}\n"


_asciigraph.graph_to_ascii = types.MethodType(
    graph_to_ascii_overload, _asciigraph
)



"""AST types module."""

from __future__ import annotations

from typing import Optional

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DataType,
    Expr,
    ReprStruct,
    SourceLocation,
)
from astx.tools.typing import typechecked


@public
@typechecked
class TypeCastExpr(Expr):
    """AST class for type casting expressions."""

    expr: Expr
    target_type: DataType

    def __init__(
        self,
        expr: Expr,
        target_type: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc=loc, parent=parent)
        self.expr = expr
        self.target_type = target_type
        self.kind = ASTKind.TypeCastExprKind

    def __str__(self) -> str:
        """Return a string representation of the TypeCast expression."""
        return f"TypeCastExpr ({self.expr}, {self.target_type})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the TypeCast expression."""
        key = "TypeCastExpr"
        value: ReprStruct = {
            "expression": self.expr.get_struct(simplified),
            "target_type": self.target_type.get_struct(simplified),
        }

        return self._prepare_struct(key, value, simplified)



"""Modifications for ASTx, such as visibility, scope, etc."""

from enum import Enum

from public import public


@public
class VisibilityKind(Enum):
    """Definition of different kind of visibility."""

    public: int = 1
    private: int = 2
    protected: int = 3


@public
class ScopeKind(Enum):
    """Definition for different kind of scopes."""

    global_: int = 1
    local: int = 2


@public
class MutabilityKind(Enum):
    """Definition for different kind of mutability."""

    constant: int = 1
    mutable: int = 2



"""Module for different kind of ASTx blocks."""

from __future__ import annotations

from typing import cast

from public import public

from astx.base import (
    ASTNodes,
    ASTType,
    ReprStruct,
)
from astx.tools.typing import typechecked


@public
@typechecked
class Block(ASTNodes[ASTType]):
    """The AST tree."""

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        block_nodes = []

        for node in self.nodes:
            block_nodes.append(node.get_struct(simplified))

        key = f"BLOCK[{self.name}]"
        value = cast(ReprStruct, block_nodes)

        return self._prepare_struct(key, value, simplified)



"""Tools for typing support helper."""

from typing import Any, Callable, TypeVar

from public import public
from typeguard import (
    CollectionCheckStrategy,
    ForwardRefPolicy,
)
from typeguard import (
    typechecked as _typechecked,
)
from typeguard._config import global_config

_T = TypeVar("_T")


__all__ = ["typechecked"]


@public
def skip_unused(*args: Any, **kwargs: Any) -> None:
    """Referencing variables to pacify static analyzers."""
    for arg in args:
        pass
    for key in kwargs:
        pass


@public
def copy_type(f: _T) -> Callable[[Any], _T]:
    """Copy types for args, kwargs from parent class."""
    skip_unused(f)
    return lambda x: x


typechecked = _typechecked(
    forward_ref_policy=ForwardRefPolicy.IGNORE,
    collection_check_strategy=CollectionCheckStrategy.ALL_ITEMS,
)

# Override the default configuration
global_config.forward_ref_policy = ForwardRefPolicy.IGNORE
global_config.collection_check_strategy = CollectionCheckStrategy.ALL_ITEMS



"""ASTx Python transpiler."""

from plum import dispatch

import astx

from astx.tools.typing import typechecked


@typechecked
class ASTxPythonTranspiler:
    """
    Transpiler that converts ASTx nodes to Python code.

    Notes
    -----
    Please keep the visit method in alphabet order according to the node type.
    The visit method for astx.AST should be the first one.
    """

    def __init__(self) -> None:
        self.indent_level = 0
        self.indent_str = "    "  # 4 spaces

    def _generate_block(self, block: astx.Block) -> str:
        """Generate code for a block of statements with proper indentation."""
        self.indent_level += 1
        indent = self.indent_str * self.indent_level
        lines = [indent + self.visit(node) for node in block.nodes]
        result = (
            "\n".join(lines)
            if lines
            else self.indent_str * self.indent_level + "pass"
        )
        self.indent_level -= 1
        return result

    @dispatch.abstract
    def visit(self, expr: astx.AST) -> str:
        """Translate an ASTx expression."""
        raise Exception(f"Not implemented yet ({expr}).")

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.AliasExpr) -> str:
        """Handle AliasExpr nodes."""
        if node.asname:
            return f"{node.name} as {node.asname}"
        return f"{node.name}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Argument) -> str:
        """Handle UnaryOp nodes."""
        type_ = self.visit(node.type_)
        return f"{node.name}: {type_}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Arguments) -> str:
        """Handle UnaryOp nodes."""
        return ", ".join([self.visit(arg) for arg in node.nodes])

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.BinaryOp) -> str:
        """Handle BinaryOp nodes."""
        lhs = self.visit(node.lhs)
        rhs = self.visit(node.rhs)
        return f"({lhs} {node.op_code} {rhs})"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Block) -> str:
        """Handle Block nodes."""
        return self._generate_block(node)

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.ClassDefStmt) -> str:
        """Handle ClassDefStmt nodes."""
        class_type = "(ABC)" if node.is_abstract else ""
        return f"class {node.name}{class_type}:" f"\n {self.visit(node.body)}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.ForRangeLoopExpr) -> str:
        """Handle ForRangeLoopExpr nodes."""
        return (
            f"result = [{self.visit(node.body)} for "
            f" {node.variable.name} in range "
            f"({self.visit(node.start)},{self.visit(node.end)},"
            f"{self.visit(node.step)})]"
        )

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Function) -> str:
        """Handle Function nodes."""
        args = self.visit(node.prototype.args)
        returns = (
            f" -> {self.visit(node.prototype.return_type)}"
            if node.prototype.return_type
            else ""
        )
        header = f"def {node.name}({args}){returns}:"
        body = self.visit(node.body)
        return f"{header}\n{body}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.FunctionReturn) -> str:
        """Handle FunctionReturn nodes."""
        value = self.visit(node.value) if node.value else ""
        return f"return {value}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.IfExpr) -> str:
        """Handle IfExpr nodes."""
        if node.else_:
            return (
                f"{self.visit(node.then)} if "
                f" {self.visit(node.condition)}"
                f" else {self.visit(node.else_)}"
            )
        return (
            f"{self.visit(node.then)} if "
            f" {self.visit(node.condition)} else None"
        )

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.IfStmt) -> str:
        """Handle IfStmt nodes."""
        if node.else_:
            return (
                f"if {self.visit(node.condition)}:"
                f"\n{self._generate_block(node.then)}"
                f"\nelse:"
                f"\n{self._generate_block(node.else_)}"
            )
        return (
            f"if {self.visit(node.condition)}:"
            f"\n{self._generate_block(node.then)}"
        )

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.ImportFromStmt) -> str:
        """Handle ImportFromStmt nodes."""
        names = [self.visit(name) for name in node.names]
        level_dots = "." * node.level
        module_str = (
            f"{level_dots}{node.module}" if node.module else level_dots
        )
        names_str = ", ".join(str(name) for name in names)
        return f"from {module_str} import {names_str}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.ImportExpr) -> str:
        """Handle ImportExpr nodes."""
        names = [self.visit(name) for name in node.names]
        names_list = []
        for name in names:
            str_ = f"__import__('{name}') "
            names_list.append(str_)
        names_str = ", ".join(x for x in names_list)

        # name if one import or name1, name2, etc if multiple imports
        num = [
            "" if len(names) == 1 else str(n) for n in range(1, len(names) + 1)
        ]
        call = ["module" + str(n) for n in num]
        call_str = ", ".join(x for x in call)

        # assign tuple if multiple imports
        names_str = (
            names_str if len(names_list) == 1 else "(" + names_str + ")"
        )

        return f"{call_str} = {names_str}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.ImportFromExpr) -> str:
        """Handle ImportFromExpr nodes."""
        names = [self.visit(name) for name in node.names]
        level_dots = "." * node.level
        module_str = (
            f"{level_dots}{node.module}" if node.module else level_dots
        )
        names_list = []
        for name in names:
            str_ = (
                f"getattr(__import__('{module_str}', "
                f"fromlist=['{name}']), '{name}')"
            )
            names_list.append(str_)
        names_str = ", ".join(x for x in names_list)

        # name if one import or name1, name2, etc if multiple imports
        num = [
            "" if len(names) == 1 else str(n) for n in range(1, len(names) + 1)
        ]
        call = ["name" + str(n) for n in num]
        call_str = ", ".join(x for x in call)

        # assign tuple if multiple imports
        names_str = (
            names_str if len(names_list) == 1 else "(" + names_str + ")"
        )

        return f"{call_str} = {names_str}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.ImportStmt) -> str:
        """Handle ImportStmt nodes."""
        names = [self.visit(name) for name in node.names]
        names_str = ", ".join(x for x in names)
        return f"import {names_str}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.ImportFromStmt) -> str:
        """Handle ImportFromStmt nodes."""
        names = [self.visit(name) for name in node.names]
        level_dots = "." * node.level
        module_str = (
            f"{level_dots}{node.module}" if node.module else level_dots
        )
        names_str = ", ".join(str(name) for name in names)
        return f"from {module_str} import {names_str}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LambdaExpr) -> str:
        """Handle LambdaExpr nodes."""
        params_str = ", ".join(param.name for param in node.params)
        return f"lambda {params_str}: {self.visit(node.body)}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralBoolean) -> str:
        """Handle LiteralBoolean nodes."""
        return "True" if node.value else "False"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralComplex32) -> str:
        """Handle LiteralComplex32 nodes."""
        real = node.value[0]
        imag = node.value[1]
        return f"complex({real}, {imag})"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralComplex) -> str:
        """Handle LiteralComplex nodes."""
        real = node.value[0]
        imag = node.value[1]
        return f"complex({real}, {imag})"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralComplex64) -> str:
        """Handle LiteralComplex64 nodes."""
        real = node.value[0]
        imag = node.value[1]
        return f"complex({real}, {imag})"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralFloat16) -> str:
        """Handle LiteralFloat nodes."""
        return str(node.value)

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralFloat32) -> str:
        """Handle LiteralFloat nodes."""
        return str(node.value)

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralFloat64) -> str:
        """Handle LiteralFloat nodes."""
        return str(node.value)

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralInt32) -> str:
        """Handle LiteralInt32 nodes."""
        return str(node.value)

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralUTF8String) -> str:
        """Handle LiteralUTF8String nodes."""
        return repr(node.value)

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralUTF8Char) -> str:
        """Handle LiteralUTF8Char nodes."""
        return repr(node.value)

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Complex32) -> str:
        """Handle Complex32 nodes."""
        return "Complex"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Complex64) -> str:
        """Handle Complex64 nodes."""
        return "Complex"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Float16) -> str:
        """Handle Float nodes."""
        return "float"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Float32) -> str:
        """Handle Float nodes."""
        return "float"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Float64) -> str:
        """Handle Float nodes."""
        return "float"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Int32) -> str:
        """Handle Int32 nodes."""
        return "int"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.TypeCastExpr) -> str:
        """Handle TypeCastExpr nodes."""
        return f"cast({self.visit(node.target_type)}, {node.expr.name})"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.UnaryOp) -> str:
        """Handle UnaryOp nodes."""
        operand = self.visit(node.operand)
        return f"({node.op_code}{operand})"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.UTF8Char) -> str:
        """Handle UTF8Char nodes."""
        return repr(node.value)

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.UTF8String) -> str:
        """Handle UTF8String nodes."""
        return repr(node.value)

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Variable) -> str:
        """Handle Variable nodes."""
        return node.name

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.VariableAssignment) -> str:
        """Handle VariableAssignment nodes."""
        target = node.name
        value = self.visit(node.value)
        return f"{target} = {value}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.WhileExpr) -> str:
        """Handle WhileExpr nodes."""
        condition = self.visit(node.condition)
        body = self.visit(node.body)
        return f"[{body} for _ in iter(lambda: {condition}, False)]"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.WhileStmt) -> str:
        """Handle WhileStmt nodes."""
        condition = self.visit(node.condition)
        body = self._generate_block(node.body)
        return f"while {condition}:\n{body}"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Date) -> str:
        """Handle Date nodes."""
        return "date"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Time) -> str:
        """Handle Time nodes."""
        return "time"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.Timestamp) -> str:
        """Handle Timestamp nodes."""
        return "timestamp"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.DateTime) -> str:
        """Handle DateTime nodes."""
        return "datetime"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralDate) -> str:
        """Handle LiteralDate nodes."""
        return f"datetime.strptime({node.value!r}, '%Y-%m-%d').date()"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralTime) -> str:
        """Handle LiteralTime nodes."""
        return f"datetime.strptime({node.value!r}, '%H:%M:%S').time()"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralTimestamp) -> str:
        """Handle LiteralTimestamp nodes."""
        return f"datetime.strptime({node.value!r}, '%Y-%m-%d %H:%M:%S')"

    @dispatch  # type: ignore[no-redef]
    def visit(self, node: astx.LiteralDateTime) -> str:
        """Handle LiteralDateTime nodes."""
        return f"datetime.strptime({node.value!r}, '%Y-%m-%dT%H:%M:%S')"



"""ASTx Data Types module."""

from __future__ import annotations

from public import public

from astx.base import (
    DataType,
)
from astx.tools.typing import typechecked


@public
@typechecked
class AnyType(DataType):
    """Generic data type expression."""



"""ASTx Data Types module."""

from __future__ import annotations

from public import public

from astx.tools.typing import typechecked
from astx.types.base import AnyType


@public
@typechecked
class Boolean(AnyType):
    """Boolean data type expression."""



"""ASTx Data Types module."""

from __future__ import annotations

from public import public

from astx.tools.typing import typechecked
from astx.types.base import AnyType


@public
@typechecked
class Temporal(AnyType):
    """Base class for temporal data types (date, time, timestamp, datetime)."""


@public
@typechecked
class Date(Temporal):
    """Date data type expression."""


@public
@typechecked
class Time(Temporal):
    """Time data type expression."""


@public
@typechecked
class Timestamp(Temporal):
    """Timestamp data type expression."""


@public
@typechecked
class DateTime(Temporal):
    """DateTime data type expression."""



"""ASTx Data Types module."""

from __future__ import annotations

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ReprStruct,
    SourceLocation,
)
from astx.literals.base import Literal
from astx.tools.typing import typechecked
from astx.types.numeric import (
    Complex,
    Complex32,
    Complex64,
    Float16,
    Float32,
    Float64,
    Int8,
    Int16,
    Int32,
    Int64,
    Int128,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    UInt128,
)


@public
@typechecked
class LiteralInt8(Literal):
    """LiteralInt8 data type class."""

    value: int

    def __init__(
        self, value: int, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralInt8."""
        super().__init__(loc)
        self.value = value
        self.type_ = Int8()
        self.loc = loc


@public
@typechecked
class LiteralInt16(Literal):
    """LiteralInt16 data type class."""

    value: int

    def __init__(
        self, value: int, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralInt16."""
        super().__init__(loc)
        self.value = value
        self.type_ = Int16()
        self.loc = loc


@public
@typechecked
class LiteralInt32(Literal):
    """LiteralInt32 data type class."""

    value: int

    def __init__(
        self, value: int, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralInt32."""
        super().__init__(loc)
        self.value = value
        self.type_ = Int32()
        self.loc = loc


@public
@typechecked
class LiteralInt64(Literal):
    """LiteralInt64 data type class."""

    value: int

    def __init__(
        self, value: int, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralInt64."""
        super().__init__(loc)
        self.value = value
        self.type_ = Int64()
        self.loc = loc


@public
@typechecked
class LiteralInt128(Literal):
    """LiteralInt128 data type class."""

    value: int

    def __init__(
        self, value: int, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralInt128."""
        super().__init__(loc)
        self.value = value
        self.type_ = Int128()
        self.loc = loc


@public
@typechecked
class LiteralUInt8(Literal):
    """LiteralUInt8 data type class."""

    value: int

    def __init__(
        self, value: int, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralUInt8."""
        super().__init__(loc)
        self.value = value
        self.type_ = UInt8()
        self.loc = loc


@public
@typechecked
class LiteralUInt16(Literal):
    """LiteralUInt16 data type class."""

    value: int

    def __init__(
        self, value: int, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralUInt16."""
        super().__init__(loc)
        self.value = value
        self.type_ = UInt16()
        self.loc = loc


@public
@typechecked
class LiteralUInt32(Literal):
    """LiteralUInt32 data type class."""

    value: int

    def __init__(
        self, value: int, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralUInt32."""
        super().__init__(loc)
        self.value = value
        self.type_ = UInt32()
        self.loc = loc


@public
@typechecked
class LiteralUInt64(Literal):
    """LiteralUInt64 data type class."""

    value: int

    def __init__(
        self, value: int, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralUInt64."""
        super().__init__(loc)
        self.value = value
        self.type_ = UInt64()
        self.loc = loc


@public
@typechecked
class LiteralUInt128(Literal):
    """LiteralUInt128 data type class."""

    value: int

    def __init__(
        self, value: int, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralUInt128."""
        super().__init__(loc)
        self.value = value
        self.type_ = UInt128()
        self.loc = loc


@public
@typechecked
class LiteralFloat16(Literal):
    """LiteralFloat16 data type class."""

    value: float

    def __init__(
        self, value: float, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralFloat16."""
        super().__init__(loc)
        self.value = value
        self.type_ = Float16()
        self.loc = loc


@public
@typechecked
class LiteralFloat32(Literal):
    """LiteralFloat32 data type class."""

    value: float

    def __init__(
        self, value: float, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralFloat32."""
        super().__init__(loc)
        self.value = value
        self.type_ = Float32()
        self.loc = loc


@public
@typechecked
class LiteralFloat64(Literal):
    """LiteralFloat64 data type class."""

    value: float

    def __init__(
        self, value: float, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralFloat64."""
        super().__init__(loc)
        self.value = value
        self.type_ = Float64()
        self.loc = loc


@public
@typechecked
class LiteralComplex(Literal):
    """Base class for literal complex numbers."""

    type_: Complex
    value: tuple[float, float]

    def __init__(
        self,
        real: float,
        imag: float,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize a generic complex number."""
        super().__init__(loc)
        self.value = real, imag

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"LiteralComplex({self.value[0]} + {self.value[1]}j)"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST representation for the complex literal."""
        key = f"{self.__class__.__name__}: {self.value}"
        value: ReprStruct = {
            "real": self.value[0],
            "imag": self.value[1],
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class LiteralComplex32(LiteralComplex):
    """LiteralComplex32 data type class."""

    def __init__(
        self,
        real: float,
        imag: float,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize LiteralComplex32."""
        super().__init__(real, imag, loc)
        self.type_ = Complex32()


@public
@typechecked
class LiteralComplex64(LiteralComplex):
    """LiteralComplex64 data type class."""

    def __init__(
        self,
        real: float,
        imag: float,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize LiteralComplex64."""
        super().__init__(real, imag, loc)
        self.type_ = Complex64()



"""Module for controle flow AST."""

from __future__ import annotations

from typing import Optional, cast

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DictDataTypesStruct,
    Expr,
    ReprStruct,
    SourceLocation,
    StatementType,
)
from astx.blocks import Block
from astx.tools.typing import typechecked
from astx.variables import InlineVariableDeclaration


@public
@typechecked
class IfStmt(StatementType):
    """AST class for `if` statement."""

    condition: Expr
    then: Block
    else_: Optional[Block]

    def __init__(
        self,
        condition: Expr,
        then: Block,
        else_: Optional[Block] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the IfStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.loc = loc
        self.condition = condition
        self.then = then
        self.else_ = else_
        self.kind = ASTKind.IfStmtKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"IfStmt[{self.condition}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        if_condition = {"condition": self.condition.get_struct(simplified)}
        if_then = {"then-block": self.then.get_struct(simplified)}
        if_else: ReprStruct = {}

        if self.else_ is not None:
            if_else = {"else-block": self.else_.get_struct(simplified)}

        key = "IF-STMT"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, if_condition),
            **cast(DictDataTypesStruct, if_then),
            **cast(DictDataTypesStruct, if_else),
        }

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class IfExpr(Expr):
    """AST class for `if` expression."""

    condition: Expr
    then: Block
    else_: Optional[Block]

    def __init__(
        self,
        condition: Expr,
        then: Block,
        else_: Optional[Block] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the IfExpr instance."""
        super().__init__(loc=loc, parent=parent)
        self.loc = loc
        self.condition = condition
        self.then = then
        self.else_ = else_
        self.kind = ASTKind.IfExprKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"IfExpr[{self.condition}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        if_condition = {"condition": self.condition.get_struct(simplified)}
        if_then = {"then-block": self.then.get_struct(simplified)}
        if_else: ReprStruct = {}

        if self.else_ is not None:
            if_else = {"else-block": self.else_.get_struct(simplified)}

        key = "IF-EXPR"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, if_condition),
            **cast(DictDataTypesStruct, if_then),
            **cast(DictDataTypesStruct, if_else),
        }

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ForRangeLoopStmt(StatementType):
    """AST class for `For` Range Statement."""

    variable: InlineVariableDeclaration
    start: Expr
    end: Expr
    step: Expr
    body: Block

    def __init__(
        self,
        variable: InlineVariableDeclaration,
        start: Expr,
        end: Expr,
        step: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ForRangeLoopStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.variable = variable
        self.start = start
        self.end = end
        self.step = step
        self.body = body
        self.kind = ASTKind.ForRangeLoopStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        start = self.start
        end = self.end
        step = self.step
        var_name = self.variable.name
        return f"ForRangeLoopStmt({var_name}=[{start}:{end}:{step}])"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_start = {"start": self.start.get_struct(simplified)}
        for_end = {"end": self.end.get_struct(simplified)}
        for_step = {"step": self.step.get_struct(simplified)}
        for_body = self.body.get_struct(simplified)

        key = "FOR-RANGE-LOOP-STMT"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, for_start),
            **cast(DictDataTypesStruct, for_end),
            **cast(DictDataTypesStruct, for_step),
            **cast(DictDataTypesStruct, for_body),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ForRangeLoopExpr(Expr):
    """AST class for `For` Range Expression."""

    variable: InlineVariableDeclaration
    start: Expr
    end: Expr
    step: Expr
    body: Block

    def __init__(
        self,
        variable: InlineVariableDeclaration,
        start: Expr,
        end: Expr,
        step: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ForRangeLoopExpr instance."""
        super().__init__(loc=loc, parent=parent)
        self.variable = variable
        self.start = start
        self.end = end
        self.step = step
        self.body = body
        self.kind = ASTKind.ForRangeLoopExprKind
        # self.step = step if step is not None else LiteralInt32(1)

    def __str__(self) -> str:
        """Return a string that represents the object."""
        var_name = self.variable.name
        # note: it would be nice to have the following structure
        #    ForRangeLoopExpr({var_name}=[{start}:{end}:{step}])
        #    but we would need to have first something like a resolver
        #    otherwise it could be a very large output
        return f"ForRangeLoopExpr[{var_name}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_var = {"var": self.variable.get_struct(simplified)}
        for_start = {"start": self.start.get_struct(simplified)}
        for_end = {"end": self.end.get_struct(simplified)}
        for_step = {"step": self.step.get_struct(simplified)}
        for_body = self.body.get_struct(simplified)

        key = "FOR-RANGE-LOOP-EXPR"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, for_var),
            **cast(DictDataTypesStruct, for_start),
            **cast(DictDataTypesStruct, for_end),
            **cast(DictDataTypesStruct, for_step),
            **cast(DictDataTypesStruct, for_body),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ForCountLoopStmt(StatementType):
    """
    AST class for a simple Count-Controlled `For` Loop statement.

    This is a very basic `for` loop, used by languages like C or C++.
    """

    initializer: InlineVariableDeclaration
    condition: Expr
    update: Expr
    body: Block

    def __init__(
        self,
        initializer: InlineVariableDeclaration,
        condition: Expr,
        update: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ForCountLoopStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.initializer = initializer
        self.condition = condition
        self.update = update
        self.body = body
        self.kind = ASTKind.ForCountLoopStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        init = self.initializer
        cond = self.condition
        update = self.update
        return f"ForCountLoopStmt({init};{cond};{update})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_init = {"initialization": self.initializer.get_struct(simplified)}
        for_cond = {"condition": self.condition.get_struct(simplified)}
        for_update = {"update": self.update.get_struct(simplified)}
        for_body = self.body.get_struct(simplified)

        key = "FOR-COUNT-STMT"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, for_init),
            **cast(DictDataTypesStruct, for_cond),
            **cast(DictDataTypesStruct, for_update),
            **cast(DictDataTypesStruct, for_body),
        }

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ForCountLoopExpr(Expr):
    """
    AST class for a simple Count-Controlled `For` Loop expression.

    This is a very basic `for` loop, used by languages like C or C++.
    """

    initializer: InlineVariableDeclaration
    condition: Expr
    update: Expr
    body: Block

    def __init__(
        self,
        initializer: InlineVariableDeclaration,
        condition: Expr,
        update: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ForLoopCountExpr instance."""
        super().__init__(loc=loc, parent=parent)
        self.initializer = initializer
        self.condition = condition
        self.update = update
        self.body = body
        self.kind = ASTKind.ForCountLoopExprKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        init = self.initializer
        cond = self.condition
        update = self.update
        return f"ForCountLoopExpr({init};{cond};{update})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_init = {"initialization": self.initializer.get_struct(simplified)}
        for_cond = {"condition": self.condition.get_struct(simplified)}
        for_update = {"update": self.update.get_struct(simplified)}
        for_body = self.body.get_struct(simplified)

        key = "FOR-COUNT-EXPR"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, for_init),
            **cast(DictDataTypesStruct, for_cond),
            **cast(DictDataTypesStruct, for_update),
            **cast(DictDataTypesStruct, for_body),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class WhileStmt(StatementType):
    """AST class for `while` statement."""

    condition: Expr
    body: Block

    def __init__(
        self,
        condition: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the WhileStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.condition = condition
        self.body = body
        self.kind = ASTKind.WhileStmtKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"WhileStmt[{self.condition}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        while_condition = self.condition.get_struct(simplified)
        while_body = self.body.get_struct(simplified)

        key = "WHILE-STMT"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, while_condition),
            **cast(DictDataTypesStruct, while_body),
        }

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class WhileExpr(Expr):
    """AST class for `while` expression."""

    condition: Expr
    body: Block

    def __init__(
        self,
        condition: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the WhileExpr instance."""
        super().__init__(loc=loc, parent=parent)
        self.condition = condition
        self.body = body
        self.kind = ASTKind.WhileExprKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"WhileExpr[{self.condition}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        while_condition = self.condition.get_struct(simplified)
        while_body = self.body.get_struct(simplified)

        key = "WHILE-EXPR"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, while_condition),
            **cast(DictDataTypesStruct, while_body),
        }

        return self._prepare_struct(key, value, simplified)



"""ASTx Data Types module."""

from __future__ import annotations

from typing import Any
from uuid import uuid4

from public import public

from astx.base import (
    ExprType,
    ReprStruct,
    SourceLocation,
)
from astx.tools.typing import typechecked
from astx.types.operators import DataTypeOps


@public
@typechecked
class Literal(DataTypeOps):
    """Literal Data type."""

    type_: ExprType
    loc: SourceLocation
    value: Any

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self.ref = uuid4().hex

    def __str__(self) -> str:
        """Return a string that represents the object."""
        klass = self.__class__.__name__
        return f"{klass}({self.value})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST representation for the object."""
        key = f"Literal[{self.type_}]: {self.value}"
        value = self.value
        return self._prepare_struct(key, value, simplified)



"""ASTx Data Types module."""

from __future__ import annotations

from public import public

from astx.tools.typing import typechecked
from astx.types.base import AnyType


@public
@typechecked
class UTF8String(AnyType):
    """Class for UTF-8 encoded strings."""


@public
@typechecked
class UTF8Char(AnyType):
    """Class for UTF-8 encoded characters."""



"""ASTx Data Types module."""

from __future__ import annotations

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    SourceLocation,
)
from astx.literals.base import Literal
from astx.tools.typing import typechecked
from astx.types.boolean import Boolean


@public
@typechecked
class LiteralBoolean(Literal):
    """LiteralBoolean data type class."""

    value: bool

    def __init__(
        self, value: bool, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralBoolean."""
        super().__init__(loc)
        self.value = value
        self.type_ = Boolean()
        self.loc = loc



"""Collection of ASTx nodes used for types."""

from astx.types.base import (
    AnyType,
)
from astx.types.boolean import (
    Boolean,
)
from astx.types.casting import (
    TypeCastExpr,
)
from astx.types.numeric import (
    Complex,
    Complex32,
    Complex64,
    Float16,
    Float32,
    Float64,
    Floating,
    Int8,
    Int16,
    Int32,
    Int64,
    Integer,
    Number,
    SignedInteger,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    UInt128,
    UnsignedInteger,
)
from astx.types.operators import (
    BinaryOp,
    DataTypeOps,
    UnaryOp,
)
from astx.types.string import (
    UTF8Char,
    UTF8String,
)
from astx.types.temporal import (
    Date,
    DateTime,
    Time,
    Timestamp,
)

__all__ = [
    "AnyType",
    "BinaryOp",
    "Boolean",
    "Complex",
    "Complex32",
    "Complex64",
    "DataTypeOps",
    "Date",
    "DateTime",
    "Float16",
    "Float32",
    "Float64",
    "Floating",
    "Int16",
    "Int32",
    "Int64",
    "Int8",
    "Integer",
    "Number",
    "SignedInteger",
    "Time",
    "Timestamp",
    "UInt128",
    "UInt16",
    "UInt32",
    "UInt64",
    "UInt8",
    "UnaryOp",
    "UTF8Char",
    "UTF8String",
    "UnsignedInteger",
    "TypeCastExpr",
]



"""ASTx Data Types module."""

from __future__ import annotations

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ReprStruct,
    SourceLocation,
)
from astx.literals.base import Literal
from astx.tools.typing import typechecked
from astx.types.string import UTF8Char, UTF8String


@public
@typechecked
class LiteralUTF8String(Literal):
    """Literal class for UTF-8 strings."""

    value: str

    def __init__(
        self, value: str, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        super().__init__(loc)
        self.value = value
        self.type_ = UTF8String()
        self.loc = loc

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"LiteralUTF8String({self.value})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the structure of the object in a simplified."""
        key = f"LiteralUTF8String: {self.value}"
        value = self.value
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class LiteralUTF8Char(Literal):
    """Literal class for UTF-8 characters."""

    value: str

    def __init__(
        self, value: str, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        super().__init__(loc)
        self.value = value
        self.type_ = UTF8Char()
        self.loc = loc

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"LiteralUTF8Char({self.value})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the structure of the object in a simplified."""
        key = f"LiteralUTF8Char: {self.value}"
        value = self.value
        return self._prepare_struct(key, value, simplified)



"""Module for classes definitions/declarations."""

from __future__ import annotations

import copy

from typing import Iterable, Optional, cast

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DictDataTypesStruct,
    Expr,
    ReprStruct,
    SourceLocation,
    StatementType,
)
from astx.blocks import Block
from astx.callables import Function
from astx.modifiers import VisibilityKind
from astx.tools.typing import typechecked
from astx.variables import VariableDeclaration


@public
@typechecked
class ClassDeclStmt(StatementType):
    """AST class for class declaration."""

    name: str
    bases: ASTNodes
    decorators: ASTNodes
    visibility: VisibilityKind
    is_abstract: bool
    metaclass: Optional[Expr]
    attributes: ASTNodes
    methods: ASTNodes

    def __init__(
        self,
        name: str,
        bases: Iterable[Expr] | ASTNodes = [],
        decorators: Iterable[Expr] | ASTNodes = [],
        visibility: VisibilityKind = VisibilityKind.public,
        is_abstract: bool = False,
        metaclass: Optional[Expr] = None,
        attributes: Iterable[VariableDeclaration] | ASTNodes = [],
        methods: Iterable[Function] | ASTNodes = [],
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize ClassDeclStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.name = name

        if isinstance(bases, ASTNodes):
            self.bases = bases
        else:
            self.bases = ASTNodes()
            for base in bases:
                self.bases.append(base)

        if isinstance(decorators, ASTNodes):
            self.decorators = decorators
        else:
            self.decorators = ASTNodes()
            for decorator in decorators:
                self.decorators.append(decorator)

        self.attributes = ASTNodes()
        for a in attributes:
            self.attributes.append(a)

        self.methods = ASTNodes()
        for m in methods:
            self.methods.append(m)

        self.visibility = visibility
        self.is_abstract = is_abstract
        self.metaclass = metaclass
        self.kind = ASTKind.ClassDeclStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        modifiers = []
        if self.visibility != VisibilityKind.public:
            modifiers.append(self.visibility.name)
        if self.is_abstract:
            modifiers.append("abstract")
        modifiers_str = " ".join(modifiers)
        bases_str = (
            ", ".join(str(base) for base in self.bases) if self.bases else ""
        )
        decorators_str = "".join(
            f"@{decorator}\n" for decorator in self.decorators
        )
        metaclass_str = (
            f" metaclass={self.metaclass}" if self.metaclass else ""
        )
        class_str = f"class {self.name}"
        if bases_str:
            class_str += f"({bases_str})"
        class_str += f"{metaclass_str}"
        return f"{decorators_str}{modifiers_str} {class_str}".strip()

    def _get_struct_wrapper(self, simplified: bool) -> DictDataTypesStruct:
        bases_dict: ReprStruct = {}
        decors_dict: ReprStruct = {}
        metaclass_dict: ReprStruct = {}
        attrs_dict: ReprStruct = {}
        methods_dict: ReprStruct = {}

        if self.bases:
            bases_dict = {"bases": self.bases.get_struct(simplified)}

        if self.decorators:
            decors_dict = {
                "decorators": self.decorators.get_struct(simplified)
            }

        if self.metaclass:
            metaclass_dict = {
                "metaclass": self.metaclass.get_struct(simplified)
            }

        if self.attributes:
            attrs_dict = {"attributes": self.attributes.get_struct(simplified)}

        if self.methods:
            methods_dict = {"methods": self.methods.get_struct(simplified)}

        value: DictDataTypesStruct = {
            **cast(DictDataTypesStruct, bases_dict),
            **cast(DictDataTypesStruct, decors_dict),
            **cast(DictDataTypesStruct, metaclass_dict),
            **cast(DictDataTypesStruct, attrs_dict),
            **cast(DictDataTypesStruct, methods_dict),
        }
        return value

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        vis = dict(zip(("public", "private", "protected"), ("+", "-", "#")))
        abstract = ", abstract" if self.is_abstract else ""

        key = f"CLASS-DECL[{vis[self.visibility.name]}{self.name}{abstract}]"
        value = self._get_struct_wrapper(simplified)

        return self._prepare_struct(key, value, simplified)


CLASS_BODY_DEFAULT = Block(name="class_body")


@public
@typechecked
class ClassDefStmt(ClassDeclStmt):
    """AST class for class definition, including attributes and methods."""

    body: Block

    def __init__(
        self,
        name: str,
        bases: Iterable[Expr] | ASTNodes = [],
        decorators: Iterable[Expr] | ASTNodes = [],
        body: Block = CLASS_BODY_DEFAULT,
        visibility: VisibilityKind = VisibilityKind.public,
        is_abstract: bool = False,
        metaclass: Optional[Expr] = None,
        attributes: Iterable[VariableDeclaration] | ASTNodes = [],
        methods: Iterable[Function] | ASTNodes = [],
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize ClassDefStmt instance."""
        super().__init__(
            name=name,
            bases=bases,
            decorators=decorators,
            visibility=visibility,
            is_abstract=is_abstract,
            metaclass=metaclass,
            attributes=attributes,
            methods=methods,
            loc=loc,
            parent=parent,
        )

        if body != CLASS_BODY_DEFAULT:
            self.body = body
        else:
            self.body = copy.deepcopy(body)
            self.body.name = f"{name}_body"
        self.kind = ASTKind.ClassDefStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        class_decl_str = super().__str__()
        if not self.body.nodes:
            body_str = "    pass"
        else:
            body_str = "\n    ".join(str(stmt) for stmt in self.body.nodes)
        return f"{class_decl_str}:\n    {body_str}"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        vis = dict(zip(("public", "private", "protected"), ("+", "-", "#")))
        abstract = ", abstract" if self.is_abstract else ""

        key = f"CLASS-DEF[{vis[self.visibility.name]}{self.name}{abstract}]"
        value = self._get_struct_wrapper(simplified)

        if self.body != CLASS_BODY_DEFAULT:
            value["body"] = self.body.get_struct(simplified)

        return self._prepare_struct(key, value, simplified)



"""AST classes and functions."""

from __future__ import annotations

import json

from abc import abstractmethod
from enum import Enum
from hashlib import sha256
from typing import (
    ClassVar,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Union,
    cast,
)

from astx.tools.typing import typechecked

try:
    from typing_extensions import TypeAlias
except ImportError:
    from typing import TypeAlias  # type: ignore[no-redef,attr-defined]


try:
    from typing_extensions import TypeVar
except ImportError:
    from typing import TypeVar  # type: ignore[assignment]


import yaml

from public import public

ASTType = TypeVar("ASTType", bound="AST", default="AST")


__all__ = [
    "ExprType",
    "PrimitivesStruct",
    "DataTypesStruct",
    "DictDataTypesStruct",
    "ReprStruct",
]


def is_using_jupyter_notebook() -> bool:
    """Check if it is executed in a jupyter notebook."""
    try:
        from IPython import get_ipython  # type: ignore

        if "IPKernelApp" in get_ipython().config:  # type: ignore
            return True
    except Exception:
        pass
    return False


@public
@typechecked
class SourceLocation:
    line: int
    col: int

    def __init__(self, line: int, col: int):
        self.line = line
        self.col = col

    def __str__(self) -> str:
        return "{" + f"line: {self.line}, col: {self.col}" + "}"

    def __repr__(self) -> str:
        return str(self)


NO_SOURCE_LOCATION = SourceLocation(-1, -1)


@public
class ASTKind(Enum):
    """The expression kind class used for downcasting."""

    GenericKind = -100
    ModuleKind = -101

    # variables
    ArgumentKind = -200
    ArgumentsKind = -201
    VariableKind = -202
    VarDeclKind = -203
    VarsDeclKind = -204
    VariableAssignmentKind = -205
    VarsAssignKind = -206

    # operators
    UnaryOpKind = -300
    BinaryOpKind = -301

    # functions
    PrototypeKind = -400
    FunctionKind = -401
    CallKind = -402
    ReturnKind = -403
    LambdaExprKind = -404

    # control flow
    IfStmtKind = -500
    ForCountLoopStmtKind = -501
    ForRangeLoopStmtKind = -502
    WhileStmtKind = -503
    ForRangeLoopExprKind = -504
    ForCountLoopExprKind = -505
    WhileExprKind = -506
    IfExprKind = -507

    # data types
    NullDTKind = -600
    BooleanDTKind = -601
    Int8DTKind = -602
    UInt8DTKind = -603
    Int16DTKind = -604
    UInt16DTKind = -605
    Int32DTKind = -606
    UInt32DTKind = -607
    Int64DTKind = -608
    UInt64DTKind = -609
    FloatDTKind = -610
    DoubleDTKind = -611
    BinaryDTKind = -612
    StringDTKind = -613
    FixedSizeBinaryDTKind = -614
    Date32DTKind = -615
    Date64DTKind = -616
    TimestampDTKind = -617
    Time32DTKind = -618
    Time64DTKind = -619
    Decimal128DTKind = -620
    Decimal256DTKind = -621
    UTF8CharDTKind = -622
    UTF8StringDTKind = -623
    TimeDTKind = -624
    DateDTKind = -625
    DateTimeDTKind = -626

    # imports(packages)
    ImportStmtKind = -700
    ImportFromStmtKind = -701
    AliasExprKind = -702
    ImportExprKind = -800
    ImportFromExprKind = -801

    TypeCastExprKind = -809

    # classes
    ClassDefStmtKind = -900
    ClassDeclStmtKind = -901


class ASTMeta(type):
    def __str__(cls) -> str:
        """Return an string that represents the object."""
        return cls.__name__


@public
@typechecked
class AST(metaclass=ASTMeta):
    """AST main expression class."""

    loc: SourceLocation
    kind: ASTKind
    comment: str
    parent: Optional[ASTNodes] = None
    ref: str

    def __init__(
        self,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the AST instance."""
        self.kind = ASTKind.GenericKind
        self.loc = loc
        self.ref = ""
        self.comment = ""
        self.parent = parent
        self._update_parent()

    def __hash__(self) -> int:
        value = sha256(f"{self.get_struct()}".encode("utf8")).digest()
        return int.from_bytes(value, "big")

    def __str__(self) -> str:
        """Return an string that represents the object."""
        return f"{self.__class__.__name__}"

    def __repr__(self) -> str:
        """Return an string that represents the object."""
        if not is_using_jupyter_notebook():
            from astx.viz import graph_to_ascii, traverse_ast_ascii

            graph = traverse_ast_ascii(self.get_struct(simplified=True))
            return graph_to_ascii(graph)
        return ""

    def _repr_png_(self) -> None:
        """
        Return PNG representation of the Graphviz object.

        This method is specially recognized by Jupyter Notebook to display
        a Graphviz diagram inline.
        """
        # importing it here in order to avoid cyclic import issue
        from astx.viz import visualize

        visualize(self.get_struct(simplified=False))

    def _update_parent(self) -> None:
        """Update the parent node."""
        if self.parent is not None:
            self.parent.append(self)

    def _get_metadata(self) -> ReprStruct:
        """Return the metadata for the requested AST."""
        metadata = {
            "loc": {"line": self.loc.line, "col": self.loc.col},
            "comment": self.comment,
            "ref": self.ref,
            "kind": self.kind.value,
        }
        return cast(ReprStruct, metadata)

    def _prepare_struct(
        self,
        key: str,
        value: Union[PrimitivesStruct, ReprStruct],
        simplified: bool,
    ) -> ReprStruct:
        struct: ReprStruct = (
            {
                key: {
                    "content": value,
                    "metadata": self._get_metadata(),
                }
            }
            if not simplified
            else {key: value}
        )
        return struct

    @abstractmethod
    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return a structure that represents the node object."""

    def to_yaml(self, simplified: bool = False) -> str:
        """Return an yaml string that represents the object."""
        return str(
            yaml.dump(self.get_struct(simplified=simplified), sort_keys=False)
        )

    def to_json(self, simplified: bool = False) -> str:
        """Return an json string that represents the object."""
        return json.dumps(self.get_struct(simplified=simplified), indent=2)


@typechecked
class ASTNodes(Generic[ASTType], AST):
    """AST with a list of nodes, supporting type-specific elements."""

    name: str
    nodes: list[ASTType]
    position: int = 0

    def __init__(
        self,
        name: str = "entry",
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the AST instance."""
        super().__init__(loc=loc, parent=parent)
        self.name = name
        self.nodes: list[ASTType] = []
        self.position: int = 0

    def __iter__(self) -> Iterator[ASTType]:
        """Overload `iter` magic function."""
        self.position = 0  # Reset position for fresh iteration
        return self

    def __next__(self) -> ASTType:
        """Overload `next` magic function."""
        if self.position >= len(self.nodes):
            self.position = 0
            raise StopIteration()

        i = self.position
        self.position += 1
        return self.nodes[i]

    def append(self, value: ASTType) -> None:
        """Append a new node to the stack."""
        self.nodes.append(value)

    def __getitem__(self, index: int) -> ASTType:
        """Support subscripting to get nodes by index."""
        return self.nodes[index]

    def __len__(self) -> int:
        """Return the number of nodes, supports len function."""
        return len(self.nodes)

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return a string that represents the object."""
        args_nodes = []

        for node in self.nodes:
            args_nodes.append(node.get_struct(simplified))

        key = str(self)
        value = cast(ReprStruct, args_nodes)
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class Expr(AST):
    """AST main expression class."""

    nbytes: int = 0


@public
@typechecked
class ExprType(Expr):
    """ExprType expression class."""

    nbytes: int = 0

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return a structure that represents the node object."""
        return {"Type": self.__class__.__name__}


@public
@typechecked
class Undefined(Expr):
    """Undefined expression class."""

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return a simple structure that represents the object."""
        value = "UNDEFINED"
        key = "UNDEFINED"
        return self._prepare_struct(key, value, simplified)


PrimitivesStruct: TypeAlias = Union[
    int,
    str,
    float,
    bool,
    Undefined,
]
DataTypesStruct: TypeAlias = Union[
    PrimitivesStruct, Dict[str, "DataTypesStruct"], List["DataTypesStruct"]
]
DictDataTypesStruct: TypeAlias = Dict[str, DataTypesStruct]
ReprStruct: TypeAlias = Union[
    List[DataTypesStruct],
    DictDataTypesStruct,
    Undefined,
]


@public
@typechecked
class DataType(ExprType):
    """AST main expression class."""

    type_: ExprType
    name: str
    _tmp_id: ClassVar[int] = 0

    def __init__(
        self,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc)
        self.name = f"temp_{DataType._tmp_id}"
        DataType._tmp_id += 1
        # set it as a generic data type
        self.type_: ExprType = ExprType()
        self.parent = parent

    def __str__(self) -> str:
        """Return an string that represents the object."""
        return f"{self.__class__.__name__}: {self.name}"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return a simple structure that represents the object."""
        key = f"DATA-TYPE[{self.__class__.__name__}]"
        value = self.name
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class OperatorType(DataType):
    """AST main expression class."""


@public
@typechecked
class StatementType(AST):
    """AST main expression class."""



"""AST nodes for literals."""

from astx.literals.base import (
    Literal,
)
from astx.literals.boolean import (
    LiteralBoolean,
)
from astx.literals.numeric import (
    LiteralComplex,
    LiteralComplex32,
    LiteralComplex64,
    LiteralFloat16,
    LiteralFloat32,
    LiteralFloat64,
    LiteralInt8,
    LiteralInt16,
    LiteralInt32,
    LiteralInt64,
    LiteralInt128,
    LiteralUInt8,
    LiteralUInt16,
    LiteralUInt32,
    LiteralUInt64,
    LiteralUInt128,
)
from astx.literals.string import (
    LiteralUTF8Char,
    LiteralUTF8String,
)
from astx.literals.temporal import (
    LiteralDate,
    LiteralDateTime,
    LiteralTime,
    LiteralTimestamp,
)

__all__ = [
    "Literal",
    "LiteralBoolean",
    "LiteralComplex",
    "LiteralComplex32",
    "LiteralComplex64",
    "LiteralFloat16",
    "LiteralFloat32",
    "LiteralFloat64",
    "LiteralInt8",
    "LiteralInt16",
    "LiteralInt32",
    "LiteralInt64",
    "LiteralInt128",
    "LiteralUInt8",
    "LiteralUInt16",
    "LiteralUInt32",
    "LiteralUInt64",
    "LiteralUInt128",
    "LiteralDate",
    "LiteralDateTime",
    "LiteralTime",
    "LiteralTimestamp",
    "LiteralUTF8Char",
    "LiteralUTF8String",
]



"""ASTx Data Types module."""

from __future__ import annotations

from public import public

from astx.tools.typing import typechecked
from astx.types.base import AnyType


@public
@typechecked
class Number(AnyType):
    """Number data type expression."""


@public
@typechecked
class Integer(AnyType):
    """Integer number data type expression."""


@public
@typechecked
class UnsignedInteger(Integer):
    """Unsigned integer number data type expression."""


@public
@typechecked
class SignedInteger(Integer):
    """Signed integer number data type expression."""


@public
@typechecked
class Int8(SignedInteger):
    """Int8 data type expression."""

    nbytes: int = 1


@public
@typechecked
class Int16(SignedInteger):
    """Int16 data type expression."""

    nbytes: int = 2


@public
@typechecked
class Int32(SignedInteger):
    """Int32 data type expression."""

    nbytes: int = 4


@public
@typechecked
class Int64(SignedInteger):
    """Int64 data type expression."""

    nbytes: int = 8


@public
@typechecked
class Int128(SignedInteger):
    """Int128 data type expression."""

    nbytes: int = 16


@public
@typechecked
class UInt8(UnsignedInteger):
    """UInt8 data type expression."""

    nbytes: int = 1


@public
@typechecked
class UInt16(UnsignedInteger):
    """UInt16 data type expression."""

    nbytes: int = 2


@public
@typechecked
class UInt32(UnsignedInteger):
    """UInt32 data type expression."""

    nbytes: int = 4


@public
@typechecked
class UInt64(UnsignedInteger):
    """UInt64 data type expression."""

    nbytes: int = 8


@public
@typechecked
class UInt128(UnsignedInteger):
    """UInt128 data type expression."""

    nbytes: int = 16


@public
@typechecked
class Floating(Number):
    """AST for the literal float number."""


@public
@typechecked
class Float16(Floating):
    """Float16 data type expression."""


@public
@typechecked
class Float32(Floating):
    """Float32 data type expression."""


@public
@typechecked
class Float64(Floating):
    """Float64 data type expression."""


@public
@typechecked
class Complex(Number):
    """Base class for complex numbers."""


@public
@typechecked
class Complex32(Complex):
    """Complex32 data type class."""

    nbytes: int = 8


@public
@typechecked
class Complex64(Complex):
    """Complex64 data type class."""

    nbytes: int = 16



"""
Symbol Table module for ASTx.

The `SymbolTable` class offered here allows the definition
of scopes, so the variable or function would be available in
specifics scopes.

"""

from __future__ import annotations

from typing import Optional, Type

from public import public

from astx.mixes import NamedExpr
from astx.tools.typing import typechecked


@public
@typechecked
class ScopeNodeBase:
    """ScopeNodeBase is the base used for the nodes (levels) in the scope."""

    name: str
    parent: Optional[ScopeNodeBase]
    default_parent: Optional[ScopeNodeBase] = None
    named_expr: dict[str, NamedExpr]

    def __init__(
        self, name: str, parent: Optional[ScopeNodeBase] = None
    ) -> None:
        """Initialize ScopeNodeBase."""
        self.parent: Optional[ScopeNodeBase] = (
            parent or ScopeNodeBase.default_parent
        )
        self.name: str = name
        self.named_expr: dict[str, NamedExpr] = {}


@public
@typechecked
class ScopeNode(ScopeNodeBase):
    """Scope node organize the scope in different levels in the stack."""

    ...


@public
@typechecked
class Scope:
    """Organize the ASTx objects according to the scope."""

    nodes: dict[int, ScopeNodeBase]
    current: Optional[ScopeNodeBase]
    previous: Optional[ScopeNodeBase]
    scope_node_class: Type[ScopeNodeBase]

    def __init__(
        self,
        scope_node_class: Type[ScopeNodeBase] = ScopeNode,
    ) -> None:
        """Initialize the scope."""
        self.nodes: dict[int, ScopeNodeBase] = {}
        self.current = None
        self.previous = None
        self.scope_node_class = scope_node_class

        self.add("root")

        self.scope_node_class.default_parent = self.current

    def add(
        self,
        name: str,
        parent: Optional[ScopeNodeBase] = None,
        change_current: bool = True,
    ) -> ScopeNodeBase:
        """Add a new node in the scope."""
        node = self.scope_node_class(name, parent)

        # The use of id(node) as keys in the nodes dictionary is generally
        # fine, but be aware that this approach may lead to potential issues
        # if the id() of a node is reused after its destruction. It's #
        # unlikely to happen in the current code, but it's something to be
        # aware of.
        self.nodes[id(node)] = node

        if len(self.nodes) == 1 or change_current:
            self.previous = self.current
            self.current = self.nodes[id(node)]

        return node

    def get_first(self) -> ScopeNodeBase:
        """Get the first node in the scope."""
        node_id = next(iter(self.nodes.keys()))
        return self.nodes[node_id]

    def get_last(self) -> ScopeNodeBase:
        """Get the latest node in the scope."""
        node_id = list(self.nodes.keys())[-1]
        return self.nodes[node_id]

    def destroy(self, node: ScopeNodeBase) -> None:
        """Destroy the current scope."""
        del self.nodes[id(node)]
        self.current = self.previous
        self.previous = None

    def set_default_parent(self, node: ScopeNodeBase) -> None:
        """Set default parent for the current scope."""
        self.scope_node_class.default_parent = node


@public
@typechecked
class SymbolTable:
    """Symbol Table for ASTx."""

    scopes: Scope

    def __init__(self) -> None:
        self.scopes = Scope()

    def define(self, expr: NamedExpr) -> None:
        """Define a new named expression inside the scoped stack."""
        if not self.scopes.current:
            raise Exception("SymbolTable: No scope active.")
        self.scopes.current.named_expr[expr.name] = expr

    def update(self, expr: NamedExpr) -> None:
        """
        Update the expression on the SymbolTable.

        It is useful mainly for updating the comment of the expression.
        """
        if not self.scopes.current:
            raise Exception("SymbolTable: No scope active.")
        if expr.name not in self.scopes.current.named_expr:
            raise Exception("This name doesn't exist in the SymbolTable.")
        self.scopes.current.named_expr[expr.name] = expr

    def lookup(self, name: str) -> NamedExpr:
        """Get a named expression from the scope stack."""
        scope = self.scopes.current
        while scope is not None:
            if name in scope.named_expr:
                return scope.named_expr[name]
            scope = scope.parent
        raise NameError(f"Name '{name}' is not defined")



"""Module for mixing ASTx types from different modules."""

from typing import Union

try:
    from typing import TypeAlias  # type: ignore
except ImportError:
    from typing_extensions import TypeAlias

from astx.base import DataType
from astx.callables import Function
from astx.variables import Variable

__all__ = ["NamedExpr"]

NamedExpr: TypeAlias = Union[DataType, Function, Variable]



"""ASTx Data Types module."""

from __future__ import annotations

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    DataType,
    ExprType,
    ReprStruct,
    SourceLocation,
)
from astx.tools.typing import typechecked


@public
@typechecked
class DataTypeOps(DataType):
    """Overload some magic functions used for the main operations."""

    def __hash__(self) -> int:
        """Ensure that the hash method is not None."""
        return super().__hash__()

    def __add__(self, other: DataType) -> BinaryOp:
        """Overload the magic `add` method."""
        return BinaryOp("+", self, other)

    def __eq__(self, other: DataType) -> BinaryOp:  # type: ignore
        """Overload the magic `eq` method."""
        return BinaryOp("==", self, other)

    def __floordiv__(self, other: DataType) -> BinaryOp:
        """Overload the magic `floordiv` method."""
        return BinaryOp("//", self, other)

    def __ge__(self, other: DataType) -> BinaryOp:
        """Overload the magic `ge` method."""
        return BinaryOp(">=", self, other)

    def __gt__(self, other: DataType) -> BinaryOp:
        """Overload the magic `gt` method."""
        return BinaryOp(">", self, other)

    def __le__(self, other: DataType) -> BinaryOp:
        """Overload the magic `le` method."""
        return BinaryOp("<=", self, other)

    def __lt__(self, other: DataType) -> BinaryOp:
        """Overload the magic `lt` method."""
        return BinaryOp("<", self, other)

    def __mod__(self, other: DataType) -> BinaryOp:
        """Overload the magic `mod` method."""
        return BinaryOp("%", self, other)

    def __mul__(self, other: DataType) -> BinaryOp:
        """Overload the magic `mul` method."""
        return BinaryOp("*", self, other)

    def __ne__(self, other: DataType) -> BinaryOp:  # type: ignore
        """Overload the magic `ne` method."""
        return BinaryOp("!=", self, other)

    def __neg__(self) -> UnaryOp:
        """Overload the magic `neg` method."""
        return UnaryOp("-", self)

    def __pos__(self) -> UnaryOp:
        """Overload the magic `pos` method."""
        return UnaryOp("+", self)

    def __pow__(self, other: DataType) -> BinaryOp:
        """Overload the magic `pow` method."""
        return BinaryOp("^", self, other)

    def __sub__(self, other: DataType) -> BinaryOp:
        """Overload the magic `sub` method."""
        return BinaryOp("-", self, other)

    def __truediv__(self, other: DataType) -> BinaryOp:
        """Overload the magic `truediv` method."""
        return BinaryOp("/", self, other)


@public
@typechecked
class UnaryOp(DataTypeOps):
    """AST class for the unary operator."""

    def __init__(
        self,
        op_code: str,
        operand: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the UnaryOp instance."""
        super().__init__()
        self.loc = loc
        self.op_code = op_code
        self.operand = operand
        self.kind = ASTKind.UnaryOpKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"UnaryOp[{self.op_code}]({self.operand})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = f"UNARY[{self.op_code}]"
        value = self.operand.get_struct(simplified)
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class BinaryOp(DataTypeOps):
    """AST class for the binary operator."""

    type_: ExprType

    def __init__(
        self,
        op_code: str,
        lhs: DataType,
        rhs: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the BinaryOp instance."""
        super().__init__()

        self.loc = loc
        self.op_code = op_code
        self.lhs = lhs
        self.rhs = rhs
        self.kind = ASTKind.BinaryOpKind

        if not (
            isinstance(lhs.type_, DataType) and isinstance(rhs.type_, DataType)
        ):
            raise Exception(
                "For now, binary operators are just allowed for `DataType`."
                f"LHS: {lhs.type_}, RHS: {rhs.type_}"
            )

        if lhs.type_ == rhs.type_:
            self.type_ = lhs.type_
        else:
            # type inference
            self.type_ = max([lhs.type_, rhs.type_], key=lambda v: v.nbytes)

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"BinaryOp[{self.op_code}]({self.lhs},{self.rhs})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure that represents the object."""
        key = f"BINARY[{self.op_code}]"
        lhs = {"lhs": self.lhs.get_struct(simplified)}
        rhs = {"rhs": self.rhs.get_struct(simplified)}

        content: ReprStruct = {**lhs, **rhs}
        return self._prepare_struct(key, content, simplified)



"""Define ASTx for more broader scope."""

from __future__ import annotations

import copy

from typing import Optional, cast

from public import public

from astx.base import (
    AST,
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    Expr,
    ReprStruct,
    SourceLocation,
    StatementType,
)
from astx.blocks import Block
from astx.tools.typing import typechecked


@public
@typechecked
class Target(Expr):
    """Define the Architecture target for the program."""

    datalayout: str
    triple: str

    def __init__(self, datalayout: str, triple: str) -> None:
        """Initialize the AST instance."""
        super().__init__()
        self.datalayout = datalayout
        self.triple = triple

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = "TARGET"
        value = f"{self.datalayout}, {self.triple}"
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class Module(Block):
    """AST main expression class."""

    name: str

    def __init__(
        self,
        name: str = "main",
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the AST instance."""
        super().__init__(name=name, loc=loc)
        self.kind = ASTKind.ModuleKind

    def __str__(self) -> str:
        """Return the string representation of the object."""
        return f"Module[{self.name}]"

    @property
    def block(self) -> list[AST]:
        """Define an alias for self.nodes."""
        return self.nodes

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        block_node = []

        for node in self.nodes:
            block_node.append(node.get_struct(simplified))

        key = f"MODULE[{self.name}]"
        value = cast(ReprStruct, block_node)

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class Package(ASTNodes):
    """AST class for Package."""

    name: str
    modules: list[Module]
    packages: list[Package]

    def __init__(
        self,
        name: str = "main",
        modules: list[Module] = [],
        packages: list[Package] = [],
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the AST instance."""
        super().__init__(loc=loc)
        self.name = name
        self.modules = copy.deepcopy(modules)
        self.packages = copy.deepcopy(packages)

    def __str__(self) -> str:
        """Return the string representation of the object."""
        return f"PACKAGE[{self.name}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        packages = []
        modules = []

        for package in self.packages:
            packages.append(package.get_struct(simplified))

        for module in self.modules:
            modules.append(module.get_struct(simplified))

        key = str(self)
        value = cast(
            ReprStruct,
            {
                "modules": modules,
                "packages": packages,
            },
        )

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class Program(Package):
    """AST class for Program."""

    target: Target

    def __init__(
        self,
        name: str = "main",
        target: Target = Target("", ""),
        modules: list[Module] = [],
        packages: list[Package] = [],
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the AST instance."""
        super().__init__(
            name=name, modules=modules, packages=packages, loc=loc
        )
        self.target = copy.deepcopy(target)

    def __str__(self) -> str:
        """Return the string representation of the object."""
        return f"PROGRAM[{self.name}]"


@public
@typechecked
class AliasExpr(Expr):
    """Represents an alias in an import statement."""

    name: str
    asname: str

    def __init__(
        self,
        name: str,
        asname: str = "",
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc=loc, parent=parent)
        self.name = name
        self.asname = asname
        self.kind = ASTKind.AliasExprKind

    def __str__(self) -> str:
        """Return a string representation of the alias."""
        if self.asname:
            return f"{self.name} as {self.asname}"
        else:
            return self.name

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the alias."""
        str_asname = f", {self.asname}" if self.asname else ""
        str_name_asname = f"[{self.name}{str_asname}]"
        key = f"Alias {str_name_asname}"
        value = ""

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ImportStmt(StatementType):
    """Represents an import statement."""

    names: list[AliasExpr]

    def __init__(
        self,
        names: list[AliasExpr],
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc=loc, parent=parent)
        self.names = names
        self.kind = ASTKind.ImportStmtKind

    def __str__(self) -> str:
        """Return a string representation of the import statement."""
        names_str = ", ".join(str(name) for name in self.names)
        return f"import {names_str}"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the import statement."""
        key = "ImportStmt"
        value = cast(
            ReprStruct, [name.get_struct(simplified) for name in self.names]
        )
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ImportFromStmt(StatementType):
    """Represents an import-from statement."""

    module: Optional[str]
    names: list[AliasExpr]
    level: int

    def __init__(
        self,
        names: list[AliasExpr],
        module: str = "",
        level: int = 0,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc=loc, parent=parent)
        self.module = module
        self.names = names
        self.level = level
        self.kind = ASTKind.ImportFromStmtKind

    def __str__(self) -> str:
        """Return a string representation of the import-from statement."""
        level_dots = "." * self.level
        module_str = (
            f"{level_dots}{self.module}" if self.module else level_dots
        )
        names_str = ", ".join(str(name) for name in self.names)
        return f"from {module_str} import {names_str}"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the import-from statement."""
        level_dots = "." * self.level
        module_str = (
            f"{level_dots}{self.module}" if self.module else level_dots
        )

        key = f"ImportFromStmt [{module_str}]"
        value = cast(
            ReprStruct, [name.get_struct(simplified) for name in self.names]
        )

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ImportExpr(Expr):
    """Represents an import operation as an expression."""

    names: list[AliasExpr]

    def __init__(
        self,
        names: list[AliasExpr],
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc=loc, parent=parent)
        self.names = names
        self.kind = ASTKind.ImportExprKind

    def __str__(self) -> str:
        """Return a string representation of the import expression."""
        names_str = ", ".join(str(name) for name in self.names)
        return f"import {names_str}"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the import expression."""
        key = "ImportExpr"
        value = cast(
            ReprStruct, [name.get_struct(simplified) for name in self.names]
        )
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ImportFromExpr(Expr):
    """Represents a 'from ... import ...' operation as an expression."""

    module: str
    names: list[AliasExpr]
    level: int  # Number of leading dots for relative imports

    def __init__(
        self,
        names: list[AliasExpr],
        module: str = "",
        level: int = 0,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc=loc, parent=parent)
        self.names = names
        self.module = module
        self.level = level
        self.kind = ASTKind.ImportFromExprKind

    def __str__(self) -> str:
        """Return a string representation of the import-from expression."""
        level_dots = "." * self.level
        module_str = (
            f"{level_dots}{self.module}" if self.module else level_dots
        )
        names_str = ", ".join(str(name) for name in self.names)

        return f"from {module_str} import {names_str}"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the import-from expression."""
        level_dots = "." * self.level
        module_str = (
            f"{level_dots}{self.module}" if self.module else level_dots
        )

        key = f"ImportFromExpr [{module_str}]"
        value = cast(
            ReprStruct, [name.get_struct(simplified) for name in self.names]
        )

        return self._prepare_struct(key, value, simplified)



"""Module for Variables."""

from __future__ import annotations

from typing import Optional

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DataType,
    Expr,
    ReprStruct,
    SourceLocation,
    StatementType,
    Undefined,
)
from astx.modifiers import MutabilityKind, ScopeKind, VisibilityKind
from astx.tools.typing import typechecked
from astx.types import AnyType
from astx.types.operators import DataTypeOps

UNDEFINED = Undefined()


@public
@typechecked
class VariableDeclaration(StatementType):
    """AST class for variable declaration."""

    mutability: MutabilityKind
    visibility: VisibilityKind
    scope: ScopeKind
    name: str
    type_: DataType
    value: Expr

    def __init__(
        self,
        name: str,
        type_: DataType,
        mutability: MutabilityKind = MutabilityKind.constant,
        visibility: VisibilityKind = VisibilityKind.public,
        scope: ScopeKind = ScopeKind.local,
        value: Expr = UNDEFINED,
        parent: Optional[ASTNodes] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the VarExprAST instance."""
        super().__init__(loc=loc, parent=parent)
        self.mutability = mutability
        self.scope = scope
        self.visibility = visibility
        self.name = name
        self.type_ = type_
        self.value = value
        self.kind = ASTKind.VarDeclKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        type_ = self.type_.__class__.__name__
        return f"VariableDeclaration[{self.name}, {type_}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return a string that represents the object."""
        key = str(self)
        value = self.value.get_struct(simplified)
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class InlineVariableDeclaration(Expr):
    """
    AST class for inline variable declaration expression.

    Can be used in expressions like for loops.
    """

    mutability: MutabilityKind
    name: str
    type_: DataType
    value: Expr

    def __init__(
        self,
        name: str,
        type_: DataType,
        mutability: MutabilityKind = MutabilityKind.constant,
        visibility: VisibilityKind = VisibilityKind.public,
        scope: ScopeKind = ScopeKind.local,
        value: Expr = UNDEFINED,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the VarExprAST instance."""
        super().__init__(loc=loc, parent=parent)
        self.mutability = mutability
        self.scope = scope
        self.visibility = visibility
        self.name = name
        self.type_ = type_
        self.value = value
        self.kind = ASTKind.VarDeclKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        type_ = self.type_.__class__.__name__
        return f"InlineVariableDeclaration[{self.name}, {type_}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return a string that represents the object."""
        key = str(self)
        value = self.value.get_struct(simplified)
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class VariableAssignment(StatementType):
    """AST class for variable declaration."""

    name: str
    value: Expr

    def __init__(
        self,
        name: str,
        value: Expr,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the VarExprAST instance."""
        super().__init__(loc=loc, parent=parent)
        self.loc = loc
        self.name = name
        self.value = value
        self.kind = ASTKind.VariableAssignmentKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"VariableAssignment[{self.name}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return a string that represents the object."""
        key = str(self)
        value = self.value.get_struct(simplified)
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class Variable(DataTypeOps):
    """AST class for the variable usage."""

    name: str
    type_: DataType = AnyType()

    def __init__(
        self,
        name: str,
        type_: DataType = AnyType(),
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the Variable instance."""
        super().__init__(loc=loc, parent=parent)
        self.name = name
        self.type_ = type_

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"Variable[{self.name}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return a string that represents the object."""
        key = f"Variable[{self.name}]"
        value = self.name
        return self._prepare_struct(key, value, simplified)



"""Module for callable ASTx."""

from __future__ import annotations

from typing import Any, Iterable, Optional, cast

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DataType,
    Expr,
    ReprStruct,
    SourceLocation,
    StatementType,
    Undefined,
)
from astx.blocks import Block
from astx.modifiers import MutabilityKind, ScopeKind, VisibilityKind
from astx.tools.typing import typechecked
from astx.types import AnyType
from astx.variables import Variable

UNDEFINED = Undefined()


@public
@typechecked
class Argument(Variable):
    """AST class for argument definition."""

    mutability: MutabilityKind
    name: str
    type_: DataType
    default: Expr

    def __init__(
        self,
        name: str,
        type_: DataType,
        mutability: MutabilityKind = MutabilityKind.constant,
        default: Expr = UNDEFINED,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the VarExprAST instance."""
        super().__init__(name=name, loc=loc, parent=parent)
        self.mutability = mutability
        self.type_ = type_
        self.default = default
        self.kind = ASTKind.ArgumentKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        type_ = self.type_.__class__.__name__
        return f"Argument[{self.name}, {type_}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = f"Argument[{self.name}, {self.type_}] = {self.default}"
        value = self.default.get_struct()
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class Arguments(ASTNodes[Argument]):
    """AST class for argument definition."""

    def __init__(self, *args: Argument, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        for arg in args:
            self.append(arg)

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"Arguments({len(self.nodes)})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        args_nodes = []

        for node in self.nodes:
            args_nodes.append(node.get_struct(simplified))

        key = str(self)
        value = cast(ReprStruct, args_nodes)
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class FunctionCall(DataType):
    """AST class for function call."""

    fn: Function
    args: Iterable[DataType]
    type_: DataType = AnyType()

    def __init__(
        self,
        fn: Function,
        args: Iterable[DataType],
        type_: DataType = AnyType(),
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the Call instance."""
        super().__init__(loc=loc, parent=parent)
        self.fn = fn
        self.args = args
        self.kind = ASTKind.CallKind
        self.type_ = type_

    def __str__(self) -> str:
        """Return a string representation of the object."""
        args = [str(arg) for arg in self.args]
        return f"Call[{self.fn}: {', '.join(args)}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        call_params = []

        for node in self.args:
            call_params.append(node.get_struct(simplified))

        key = f"FUNCTION-CALL[{self.fn.name}]"
        value = cast(
            ReprStruct,
            {
                f"Parameters ({len(call_params)})": {
                    f"param({idx})": param
                    for idx, param in enumerate(call_params)
                }
            },
        )

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class FunctionPrototype(StatementType):
    """AST class for function prototype declaration."""

    name: str
    args: Arguments
    return_type: AnyType
    scope: ScopeKind
    visibility: VisibilityKind

    def __init__(
        self,
        name: str,
        args: Arguments,
        return_type: AnyType,
        scope: ScopeKind = ScopeKind.global_,
        visibility: VisibilityKind = VisibilityKind.public,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the FunctionPrototype instance."""
        super().__init__(loc=loc, parent=parent)
        self.name = name
        self.args = args
        self.return_type = return_type
        self.loc = loc
        self.kind = ASTKind.PrototypeKind
        self.scope = scope
        self.visibility = visibility

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Get the AST structure that represent the object."""
        raise Exception("Visitor method not necessary")


@public
@typechecked
class FunctionReturn(StatementType):
    """AST class for function `return` statement."""

    value: DataType

    def __init__(
        self,
        value: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the Return instance."""
        super().__init__(loc=loc, parent=parent)
        self.value = value
        self.kind = ASTKind.ReturnKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"Return[{self.value}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = "RETURN"
        value = self.value.get_struct(simplified)
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class Function(StatementType):
    """AST class for function definition."""

    prototype: FunctionPrototype
    body: Block

    def __init__(
        self,
        prototype: FunctionPrototype,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the Function instance."""
        super().__init__(loc=loc, parent=parent)
        self.prototype = prototype
        self.body = body
        self.kind = ASTKind.FunctionKind

    @property
    def name(self) -> str:
        """Return the function prototype name."""
        return self.prototype.name

    def __str__(self) -> str:
        """Return a string that represent the object."""
        return f"Function[{self.name}]"

    def __call__(
        self,
        args: tuple[DataType, ...],
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> FunctionCall:
        """Initialize the Call instance."""
        return FunctionCall(fn=self, args=args, loc=loc, parent=parent)

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Get the AST structure that represent the object."""
        fn_args = self.prototype.args.get_struct(simplified)
        fn_body = self.body.get_struct(simplified)

        key = f"FUNCTION[{self.prototype.name}]"
        args_struct = {"args": fn_args}
        body_struct = {"body": fn_body}

        value: ReprStruct = {**args_struct, **body_struct}
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class LambdaExpr(Expr):
    """AST class for lambda expressions."""

    params: Arguments = Arguments()
    body: Expr

    def __init__(
        self,
        body: Expr,
        params: Arguments = Arguments(),
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc=loc, parent=parent)
        self.params = params
        self.body = body
        self.kind = ASTKind.LambdaExprKind

    def __str__(self) -> str:
        """Return a string representation of the lambda expression."""
        params_str = ", ".join(param.name for param in self.params)
        return f"lambda {params_str}: {self.body}"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the lambda expression."""
        key = "LambdaExpr"
        value: ReprStruct = {
            "params": self.params.get_struct(simplified),
            "body": self.body.get_struct(simplified),
        }
        return self._prepare_struct(key, value, simplified)



"""ASTx Data Types module."""

from __future__ import annotations

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ReprStruct,
    SourceLocation,
)
from astx.literals.base import Literal
from astx.tools.typing import typechecked
from astx.types.temporal import (
    Date,
    DateTime,
    Time,
    Timestamp,
)


@public
@typechecked
class LiteralDate(Literal):
    """LiteralDate data type class."""

    def __init__(
        self, value: str, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralDate."""
        super().__init__(loc)
        self.value = value
        self.type_ = Date()
        self.loc = loc

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"LiteralDate[{self.value}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the structure of the LiteralDate object."""
        key = f"LiteralDate: {self.value}"
        return self._prepare_struct(key, self.value, simplified)


@public
@typechecked
class LiteralTime(Literal):
    """LiteralTime data type class."""

    def __init__(
        self, value: str, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralTime."""
        super().__init__(loc)
        self.value = value
        self.type_ = Time()
        self.loc = loc

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"LiteralTime[{self.value}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the structure of the LiteralTime object."""
        key = f"LiteralTime: {self.value}"
        return self._prepare_struct(key, self.value, simplified)


@public
@typechecked
class LiteralTimestamp(Literal):
    """LiteralTimestamp data type class."""

    def __init__(
        self, value: str, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralTimestamp."""
        super().__init__(loc)
        self.value = value
        self.type_ = Timestamp()
        self.loc = loc

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"LiteralTimestamp[{self.value}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the structure of the LiteralTimestamp object."""
        key = f"LiteralTimestamp: {self.value}"
        return self._prepare_struct(key, self.value, simplified)


@public
@typechecked
class LiteralDateTime(Literal):
    """LiteralDateTime data type class."""

    def __init__(
        self, value: str, loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralDateTime."""
        super().__init__(loc)
        self.value = value
        self.type_ = DateTime()
        self.loc = loc

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"LiteralDateTime[{self.value}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the structure of the LiteralDateTime object."""
        key = f"LiteralDateTime: {self.value}"
        return self._prepare_struct(key, self.value, simplified)



# mypy: disable-error-code="attr-defined"
"""ASTx."""

from importlib import metadata as importlib_metadata

from astx import (
    base,
    blocks,
    callables,
    flows,
    literals,
    mixes,
    packages,
    symbol_table,
    types,
    variables,
)
from astx.base import (
    AST,
    ASTKind,
    DataType,
    Expr,
    ExprType,
    OperatorType,
    SourceLocation,
    StatementType,
    Undefined,
)
from astx.blocks import (
    Block,
)
from astx.callables import (
    Argument,
    Arguments,
    Function,
    FunctionCall,
    FunctionPrototype,
    FunctionReturn,
    LambdaExpr,
)
from astx.classes import (
    ClassDeclStmt,
    ClassDefStmt,
)
from astx.flows import (
    ForCountLoopExpr,
    ForCountLoopStmt,
    ForRangeLoopExpr,
    ForRangeLoopStmt,
    IfExpr,
    IfStmt,
    WhileExpr,
    WhileStmt,
)
from astx.literals import (
    Literal,
    LiteralBoolean,
    LiteralComplex,
    LiteralComplex32,
    LiteralComplex64,
    LiteralDate,
    LiteralDateTime,
    LiteralFloat16,
    LiteralFloat32,
    LiteralFloat64,
    LiteralInt8,
    LiteralInt16,
    LiteralInt32,
    LiteralInt64,
    LiteralInt128,
    LiteralTime,
    LiteralTimestamp,
    LiteralUInt8,
    LiteralUInt16,
    LiteralUInt32,
    LiteralUInt64,
    LiteralUInt128,
    LiteralUTF8Char,
    LiteralUTF8String,
)
from astx.mixes import (
    NamedExpr,
)
from astx.modifiers import (
    MutabilityKind,
    ScopeKind,
    VisibilityKind,
)
from astx.packages import (
    AliasExpr,
    ImportExpr,
    ImportFromExpr,
    ImportFromStmt,
    ImportStmt,
    Module,
    Package,
    Program,
    Target,
)
from astx.types import (
    BinaryOp,
    Boolean,
    Complex,
    Complex32,
    Complex64,
    DataTypeOps,
    Date,
    DateTime,
    Float16,
    Float32,
    Float64,
    Floating,
    Int8,
    Int16,
    Int32,
    Int64,
    Integer,
    Number,
    SignedInteger,
    Time,
    Timestamp,
    TypeCastExpr,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    UInt128,
    UnaryOp,
    UnsignedInteger,
    UTF8Char,
    UTF8String,
)
from astx.variables import (
    InlineVariableDeclaration,
    Variable,
    VariableAssignment,
    VariableDeclaration,
)


def get_version() -> str:
    """Return the program version."""
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "0.17.0"  # semantic-release


__all__ = [
    "AliasExpr",
    "Argument",
    "Arguments",
    "AST",
    "ASTKind",
    "base",
    "BinaryOp",
    "Block",
    "blocks",
    "Boolean",
    "callables",
    "DataType",
    "DataTypeOps",
    "datatypes",
    "Expr",
    "ExprType",
    "Float16",
    "Float32",
    "Float64",
    "LiteralFloat16",
    "LiteralFloat32",
    "LiteralFloat64",
    "Floating",
    "flows",
    "ForCountLoopStmt",
    "ForCountLoopExpr",
    "ForRangeLoopStmt",
    "ForRangeLoopExpr",
    "Function",
    "FunctionCall",
    "FunctionPrototype",
    "FunctionReturn",
    "get_version",
    "IfStmt",
    "IfExpr",
    "ImportFromExpr",
    "ImportExpr",
    "ImportStmt",
    "ImportFromStmt",
    "InlineVariableDeclaration",
    "Int16",
    "Int32",
    "Int64",
    "Int8",
    "Integer",
    "UnsignedInteger",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "UInt128",
    "LambdaExpr",
    "Literal",
    "LiteralBoolean",
    "LiteralInt8",
    "LiteralInt16",
    "LiteralInt32",
    "LiteralInt64",
    "LiteralInt128",
    "LiteralUInt8",
    "LiteralUInt16",
    "LiteralUInt32",
    "LiteralUInt64",
    "LiteralUInt128",
    "mixes",
    "Module",
    "MutabilityKind",
    "NamedExpr",
    "Number",
    "OperatorType",
    "packages",
    "Package",
    "Program",
    "ScopeKind",
    "SignedInteger",
    "SourceLocation",
    "StatementType",
    "symbol_table",
    "Target",
    "TypeCastExpr",
    "UnaryOp",
    "Undefined",
    "VariableAssignment",
    "VariableDeclaration",
    "Variable",
    "variables",
    "VisibilityKind",
    "WhileStmt",
    "WhileExpr",
    "Complex",
    "Complex32",
    "Complex64",
    "ClassDefStmt",
    "ClassDeclStmt",
    "LiteralComplex",
    "LiteralComplex32",
    "LiteralComplex64",
    "LiteralUTF8Char",
    "LiteralUTF8String",
    "UTF8Char",
    "UTF8String",
    "Date",
    "DateTime",
    "Timestamp",
    "Time",
    "LiteralDate",
    "LiteralDateTime",
    "LiteralTimestamp",
    "LiteralTime",
    "types",
    "literals",
]


version: str = get_version()

__author__ = "Ivan Ogasawara"
__email__ = "ivan.ogasawara@gmail.com"
__version__: str = version
